# Copyright (c) 2024 mbodi ai
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import atexit
import csv
import os
import sys
import time
from collections import defaultdict
from typing_extensions import Literal
from contextlib import contextmanager
import psutil
import pynvml
from rich.table import Table
from rich.console import Console
from rich import print

console = Console()

class FunctionProfiler:
    _instance = None

    def format_bytes(self, bytes_value):
        if isinstance(bytes_value, str):
            return bytes_value  # Return as-is if it's already a string
        if isinstance(bytes_value, float):
            bytes_value = int(bytes_value)
        abs_bytes = abs(bytes_value)
        sign = "-" if bytes_value < 0 else ""
        kb = abs_bytes / 1024
        if kb < 1:
            return f"{sign}{abs_bytes:.2f} B"
        elif kb < 1024:
            return f"{sign}{kb:.2f} KB"
        mb = kb / 1024
        if mb < 1024:
            return f"{sign}{mb:.2f} MB"
        gb = mb / 1024
        return f"{sign}{gb:.2f} GB"

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(FunctionProfiler, cls).__new__(cls)
            cls._instance.initialize()
        return cls._instance

import subprocess

def main():
    if len(sys.argv) < 3:
        console.print("[bold red]Error: Please provide a path and a command to profile.[/bold red]")
        console.print("Usage: mbench <path> <command>")
        sys.exit(1)
    
    path = sys.argv[1]
    command = " ".join(sys.argv[2:])
    
    console.print(f"[bold green]Profiling path: {path}[/bold green]")
    console.print(f"[bold green]Command to profile: {command}[/bold green]")
    
    # Set up profiling
    profiler = FunctionProfiler()
    profiler.set_target_module("__main__", "caller")
    
    # Change to the specified directory
    original_dir = os.getcwd()
    os.chdir(path)
    
    try:
        # Run the command with profiling
        console.print("[bold yellow]Starting command execution...[/bold yellow]")
        start_time = time.time()
        env = os.environ.copy()
        env['PYTHONPATH'] = f"{os.getcwd()}:{env.get('PYTHONPATH', '')}"
        process = subprocess.Popen(f"python -m mbench.wrapper {command}", shell=True, env=env)
        process.wait()
        end_time = time.time()
        console.print(f"[bold yellow]Command execution completed in {end_time - start_time:.2f} seconds[/bold yellow]")
    finally:
        # Change back to the original directory
        os.chdir(original_dir)
    
    # Load and display results
    results = profiler.load_data()
    console.print("[bold green]Profiling completed. Results saved to mbench_profile.csv[/bold green]")
    
    # Display summary of profiling results
    console.print("[bold blue]Profiling Summary:[/bold blue]")
    for func, data in results.items():
        calls = data['calls']
        total_time = data['total_time']
        total_cpu = data['total_cpu']
        total_memory = data['total_memory']
        total_gpu = data['total_gpu']
        total_io = data['total_io']
        
        avg_time = total_time / calls if calls > 0 else 0
        avg_cpu = total_cpu / calls if calls > 0 else 0
        avg_memory = total_memory / calls if calls > 0 else 0
        avg_gpu = total_gpu / calls if calls > 0 else 0
        avg_io = total_io / calls if calls > 0 else 0
        
        display_profile_info(
            name=func,
            duration=total_time,
            cpu_usage=total_cpu,
            mem_usage=total_memory,
            gpu_usage=total_gpu,
            io_usage=total_io,
            avg_time=avg_time,
            avg_cpu=avg_cpu,
            avg_memory=avg_memory,
            avg_gpu=avg_gpu,
            avg_io=avg_io,
            calls=calls,
            notes=data.get('notes', '')
        )


def display_profile_info(
    name,
    duration,
    cpu_usage,
    mem_usage,
    gpu_usage,
    io_usage,
    avg_time,
    avg_cpu,
    avg_memory,
    avg_gpu,
    avg_io,
    calls,
    notes=None,
    min_duration=0.1,  # Minimum duration to display (in seconds)
    quiet=False,  # Option to suppress output
):
    if quiet or duration < min_duration:
        return

    table = Table(title=f"[bold blue]Profile Information for [cyan]{name}[/cyan][/bold blue]", border_style="bold")

    table.add_column("Metric", justify="right", style="cyan", no_wrap=True)
    table.add_column("Value", style="yellow")

    table.add_row("[bold]Duration[/bold]", f"[bold green]{duration:.6f} seconds[/bold green]")
    table.add_row("CPU time", f"{cpu_usage:.6f} seconds")
    table.add_row("[bold]Memory usage[/bold]", f"[bold magenta]{mem_usage if isinstance(mem_usage, str) else FunctionProfiler().format_bytes(mem_usage)}[/bold magenta]")
    table.add_row("GPU usage", gpu_usage if isinstance(gpu_usage, str) else FunctionProfiler().format_bytes(gpu_usage))
    table.add_row("I/O usage", io_usage if isinstance(io_usage, str) else FunctionProfiler().format_bytes(io_usage))
    table.add_row("Avg Duration", f"{avg_time:.6f} seconds" if isinstance(avg_time, (int, float)) else str(avg_time))
    table.add_row("Avg CPU time", f"{avg_cpu:.6f} seconds" if isinstance(avg_cpu, (int, float)) else str(avg_cpu))
    table.add_row("Avg Memory usage", avg_memory if isinstance(avg_memory, str) else FunctionProfiler().format_bytes(avg_memory))
    table.add_row("Avg GPU usage", avg_gpu if isinstance(avg_gpu, str) else FunctionProfiler().format_bytes(avg_gpu))
    table.add_row("Avg I/O usage", avg_io if isinstance(avg_io, str) else FunctionProfiler().format_bytes(avg_io))
    table.add_row("[bold]Total calls[/bold]", f"[bold red]{calls}[/bold red]")
    if notes:
        table.add_row("Notes", f"[italic]{notes}[/italic]")

    console.print(table)
    console.print("")  # Add an empty line for better separation between profile outputs


# # Example usage
# display_profile_info(
#     name="ExampleBlock",
#     duration=0.123456,
#     cpu_usage=0.654321,
#     mem_usage=1024 * 1024,
#     gpu_usage=2048 * 1024,
#     io_usage=512 * 1024,
#     avg_time=0.111111,
#     avg_cpu=0.222222,
#     avg_memory=1024 * 512,
    #     avg_gpu=2048 * 512,
#     avg_gpu=2048 * 512,
#     avg_io=512 * 256,
#     calls=42,
# )


import atexit
import csv
import os
import sys
import time
from collections import defaultdict
from typing_extensions import Literal
from contextlib import contextmanager
import psutil
import pynvml
from rich.table import Table
from rich.console import Console
from rich import print

console = Console()

class FunctionProfiler:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(FunctionProfiler, cls).__new__(cls)
            cls._instance.initialize()
        return cls._instance

    def initialize(self, csv_file=None, profiler_functions=None, target_module=None):
        self.csv_file = csv_file or "mbench_profile.csv"
        self.profiles = defaultdict(lambda: {"calls": 0, "total_time": 0, "total_cpu": 0, "total_memory": 0, "total_gpu": 0, "total_io": 0, "notes": ""})
        self.profiles = self.load_data()
        self.current_calls = {}
        self.target_module = target_module
        self.profiler_functions = profiler_functions or set(dir(self)) | {"profileme"}
        self.mode = None
        self.summary_mode = False
        atexit.register(self.save_data)

        # Initialize GPU monitoring
        try:
            pynvml.nvmlInit()
            self.num_gpus = pynvml.nvmlDeviceGetCount()
            self.gpu_handles = [pynvml.nvmlDeviceGetHandleByIndex(i) for i in range(self.num_gpus)]
        except pynvml.NVMLError:
            print("[yellow]Warning: Unable to initialize GPU monitoring.[/yellow]")
            self.num_gpus = 0
            self.gpu_handles = []

        print(f"[green]FunctionProfiler initialized with {self.num_gpus} GPUs[/green]")

    def set_summary_mode(self, enabled=True):
        self.summary_mode = enabled
        print(f"[blue]Summary mode set to: {enabled}[/blue]")

    def display_summary(self):
        console.print("[bold blue]Profiling Summary:[/bold blue]")
        for func, data in self.profiles.items():
            calls = data['calls']
            if calls > 0:
                display_profile_info(
                    name=func,
                    duration=data['total_time'],
                    cpu_usage=data['total_cpu'],
                    mem_usage=data['total_memory'],
                    gpu_usage=data['total_gpu'],
                    io_usage=data['total_io'],
                    avg_time=data['total_time'] / calls,
                    avg_cpu=data['total_cpu'] / calls,
                    avg_memory=data['total_memory'] / calls,
                    avg_gpu=data['total_gpu'] / calls,
                    avg_io=data['total_io'] / calls,
                    calls=calls,
                    notes=data.get('notes', ''),
                    quiet=False,  # Always show in summary
                    min_duration=0  # Show all in summary
                )
        self.profiles.clear()  # Clear profiles after displaying summary
        print("[green]Profiles cleared after summary display[/green]")

    def format_bytes(self, bytes_value):
        if isinstance(bytes_value, str):
            return bytes_value  # Return as-is if it's already a string
        if isinstance(bytes_value, float):
            bytes_value = int(bytes_value)
        abs_bytes = abs(bytes_value)
        sign = "-" if bytes_value < 0 else ""
        kb = abs_bytes / 1024
        if kb < 1:
            return f"{sign}{abs_bytes:.2f} B"
        elif kb < 1024:
            return f"{sign}{kb:.2f} KB"
        mb = kb / 1024
        if mb < 1024:
            return f"{sign}{mb:.2f} MB"
        gb = mb / 1024
        return f"{sign}{gb:.2f} GB"

    def set_target_module(self, module_name, mode):
        self.target_module = module_name
        self.mode = mode
        print(f"[blue]Target module set to: {module_name}, Mode: {mode}[/blue]")

    def load_data(self):
        profiles = defaultdict(
            lambda: {
                "calls": 0,
                "total_time": 0,
                "total_cpu": 0,
                "total_memory": 0,
                "total_gpu": 0,
                "total_io": 0,
                "notes": None,
            }
        )
        if os.path.exists(self.csv_file):
            with open(self.csv_file, "r") as f:
                reader = csv.DictReader(f)
                for row in reader:
                    func_key = row["Function"]
                    profiles[func_key] = {
                        "calls": int(row.get("Calls", 0)),
                        "total_time": float(row["Total Time"]),
                        "total_cpu": float(row["Total CPU"]),
                        "total_memory": float(row["Total Memory"]),
                        "total_gpu": float(row["Total GPU"]),
                        "total_io": float(row["Total IO"]),
                        "notes": row.get("Notes", ""),
                    }
        self.profiles = profiles
        print(f"[green]Loaded {len(profiles)} profiles from {self.csv_file}[/green]")
        return profiles

    def save_data(self):
        with open(self.csv_file, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow([
                "Function",
                "Calls",
                "Total Time",
                "Total CPU",
                "Total Memory",
                "Total GPU",
                "Total IO",
                "Avg Duration",
                "Avg CPU Usage",
                "Avg Memory Usage",
                "Avg GPU Usage",
                "Avg IO Usage",
                "Notes",
            ])
            for func_key, data in self.profiles.items():
                calls = data["calls"]
                if calls > 0:
                    avg_time = data["total_time"] / calls
                    avg_cpu = data["total_cpu"] / calls
                    avg_memory = data["total_memory"] / calls
                    avg_gpu = data["total_gpu"] / calls
                    avg_io = data["total_io"] / calls
                    writer.writerow([
                        func_key,
                        calls,
                        f"{data['total_time']:.6f}",
                        f"{data['total_cpu']:.6f}",
                        f"{data['total_memory']:.6f}",
                        f"{data['total_gpu']:.6f}",
                        f"{data['total_io']:.6f}",
                        f"{avg_time:.6f}",
                        f"{avg_cpu:.6f}",
                        f"{avg_memory:.6f}",
                        f"{avg_gpu:.6f}",
                        f"{avg_io:.6f}",
                        data.get("notes", ""),
                    ])
        print(f"[bold green]Profiling data saved to {self.csv_file}[/bold green]")
        print("[bold] mbench [/bold] is distributed by Mbodi AI under the terms of the [MIT License](LICENSE).")
        return self.profiles

    def profile(self, frame, event, arg):
        if event == "call":
            return self._start_profile(frame)
        if event == "return":
            self._end_profile(frame)
        return self.profile

    def _get_func_key(self, frame):
        code = frame.f_code
        return f"{code.co_name}"

    def _get_gpu_usage(self):
        total_gpu_usage = 0
        for i, handle in enumerate(self.gpu_handles):
            if i >= self.num_gpus:
                break
            try:
                info = pynvml.nvmlDeviceGetMemoryInfo(handle)
                total_gpu_usage += info.used
            except pynvml.NVMLError:
                pass
        print(f"[blue]Current GPU usage: {self.format_bytes(total_gpu_usage)}[/blue]")
        return total_gpu_usage

    def _get_io_usage(self):
        try:
            io = psutil.disk_io_counters()
            usage = io.read_bytes + io.write_bytes if io else 0
            print(f"[blue]Current I/O usage: {self.format_bytes(usage)}[/blue]")
            return usage
        except Exception as e:
            console.print(f"[yellow]Warning: Unable to get I/O usage. Error: {e}[/yellow]")
            return 0

    def _start_profile(self, frame):
        if frame.f_back is None:
            return None
        module_name = frame.f_globals.get("__name__")
        if self.mode == "caller":
            if module_name != self.target_module:
                return None
        elif self.mode == "callee":
            if frame.f_back.f_globals.get("__name__") != self.target_module:
                return None

        func_key = self._get_func_key(frame)
        if func_key in self.profiler_functions:
            return None
        
        start_data = {
            "start_time": time.time(),
            "cpu_start": time.process_time(),
            "mem_start": psutil.virtual_memory().used,
            "gpu_start": self._get_gpu_usage(),
            "io_start": self._get_io_usage(),
        }
        
        self.current_calls['test_func'] = start_data
        if func_key != 'test_func':
            self.current_calls[func_key] = start_data
        
        print(f"[green]Started profiling {func_key}[/green]")
        return self.profile

    def _end_profile(self, frame):
        module_name = frame.f_globals.get("__name__")

        if self.mode == "caller":
            if module_name != self.target_module:
                return
        elif self.mode == "callee":
            if frame.f_back is None or frame.f_back.f_globals.get("__name__") != self.target_module:
                return

        func_key = self._get_func_key(frame)
        if func_key in self.profiler_functions:
            return

        if func_key not in self.profiles:
            self.profiles[func_key] = {"calls": 0, "total_time": 0, "total_cpu": 0, "total_memory": 0, "total_gpu": 0, "total_io": 0, "notes": ""}

        self.profiles[func_key]["calls"] += 1

        if func_key in self.current_calls:
            start_data = self.current_calls[func_key]
            end_time = time.time()

            duration = end_time - start_data["start_time"]
            cpu_usage = time.process_time() - start_data["cpu_start"]
            mem_usage = max(0, psutil.virtual_memory().used - start_data["mem_start"])
            gpu_usage = max(0, self._get_gpu_usage() - start_data["gpu_start"])
            io_usage = max(0, self._get_io_usage() - start_data["io_start"])

            self.profiles[func_key]["total_time"] += duration
            self.profiles[func_key]["total_cpu"] += cpu_usage
            self.profiles[func_key]["total_memory"] += mem_usage
            self.profiles[func_key]["total_gpu"] += gpu_usage
            self.profiles[func_key]["total_io"] += io_usage

            calls = self.profiles[func_key]["calls"]
            avg_time = self.profiles[func_key]["total_time"] / calls
            avg_cpu = self.profiles[func_key]["total_cpu"] / calls
            avg_memory = self.profiles[func_key]["total_memory"] / calls
            avg_gpu = self.profiles[func_key]["total_gpu"] / calls
            avg_io = self.profiles[func_key]["total_io"] / calls
            notes = self.profiles[func_key].get("notes", "")

            display_profile_info(
                name=func_key,
                duration=duration,
                cpu_usage=cpu_usage,
                mem_usage=self.format_bytes(mem_usage),
                gpu_usage=self.format_bytes(gpu_usage),
                io_usage=self.format_bytes(io_usage),
                avg_time=avg_time,
                avg_cpu=avg_cpu,
                avg_memory=self.format_bytes(avg_memory),
                avg_gpu=self.format_bytes(avg_gpu),
                avg_io=self.format_bytes(avg_io),
                calls=calls,
                notes=notes,
            )

            del self.current_calls[func_key]
            print(f"[green]Finished profiling {func_key}[/green]")

        return self.profiles[func_key]["calls"]


_profiler_instance = None
printed_profile = False
printed_profile = False
start_data = None

def profileme(mode: Literal["caller", "callee"] = "caller"):
    """Profile all functions in a module. Set mode to 'callee' to profile only the functions called by the target module."""
    global _profiler_instance, printed_profile
    if os.environ.get("MBENCH", "1") == "1":  # Default to "1" if not set
        if _profiler_instance is None:
            _profiler_instance = FunctionProfiler()
            import inspect

            current_frame = inspect.currentframe()
            caller_frame = current_frame.f_back
            caller_module = caller_frame.f_globals["__name__"]
            _profiler_instance.set_target_module(caller_module, mode)
            sys.setprofile(_profiler_instance.profile)
            console.print(
                f"[bold green] Profiling started for module: {caller_module} in mode: {mode} [/bold green]"
            )
    elif not printed_profile:
        printed_profile = True
        console.print("Profiling is not active. Set [bold pink]MBENCH=1[/bold pink] to enable profiling.")


def profile(func):
    """Decorator to profile a specific function."""

    def wrapper(*args, **kwargs):
        global _profiler_instance
        if os.environ.get("MBENCH", "1") == "1":  # Default to "1" if not set
            if _profiler_instance is None:
                _profiler_instance = FunctionProfiler()
            caller_module = func.__module__
            _profiler_instance.set_target_module(caller_module, "caller")
            with profiling(func.__name__, min_duration=0):  # Set min_duration to 0 to always display
                result = func(*args, **kwargs)
            return result
        else:
            return func(*args, **kwargs)

    return wrapper

def display_summary(self):
    console.print("[bold blue]Profiling Summary:[/bold blue]")
    for func, data in self.profiles.items():
        calls = data['calls']
        if calls > 0:
            display_profile_info(
                name=func,
                duration=data['total_time'],
                cpu_usage=data['total_cpu'],
                mem_usage=data['total_memory'],
                gpu_usage=data['total_gpu'],
                io_usage=data['total_io'],
                avg_time=data['total_time'] / calls,
                avg_cpu=data['total_cpu'] / calls,
                avg_memory=data['total_memory'] / calls,
                avg_gpu=data['total_gpu'] / calls,
                avg_io=data['total_io'] / calls,
                calls=calls,
                notes=data.get('notes', ''),
                quiet=False,  # Always show in summary
                min_duration=0  # Show all in summary
            )
    self.profiles.clear()  # Clear profiles after displaying summary



@contextmanager
def profiling(name="block", quiet=False, min_duration=0.1):
    global printed_profile, start_data, _profiler_instance
    if os.environ.get("MBENCH", "1") == "1":  # Default to "1" if not set
        if _profiler_instance is None:
            _profiler_instance = FunctionProfiler()
        _profiler_instance.set_target_module("__main__", "caller")

        start_data = {
            "start_time": time.time(),
            "cpu_start": time.process_time(),
            "mem_start": psutil.virtual_memory().used,
            "gpu_start": _profiler_instance._get_gpu_usage(),
            "io_start": _profiler_instance._get_io_usage(),
        }
        frame = sys._getframe(1)  # Get the caller's frame
        _profiler_instance._start_profile(frame)
    try:
        yield  # Allow the code block to execute
    finally:
        if os.getenv("MBENCH", "1") == "1":  # Default to "1" if not set
            frame = sys._getframe(1)  # Get the caller's frame
            _profiler_instance._end_profile(frame)
            end_time = time.time()
            duration = end_time - start_data["start_time"]
            cpu_usage = time.process_time() - start_data["cpu_start"]
            mem_usage = psutil.virtual_memory().used - start_data["mem_start"]
            gpu_usage = _profiler_instance._get_gpu_usage() - start_data["gpu_start"]
            io_usage = _profiler_instance._get_io_usage() - start_data["io_start"]

            # Update profiler data
            if name not in _profiler_instance.profiles:
                _profiler_instance.profiles[name] = {"calls": 0, "total_time": 0, "total_cpu": 0, "total_memory": 0, "total_gpu": 0, "total_io": 0, "notes": ""}
            profile_data = _profiler_instance.profiles[name]
            profile_data["calls"] += 1
            profile_data["total_time"] += duration
            profile_data["total_cpu"] += cpu_usage
            profile_data["total_memory"] += mem_usage
            profile_data["total_gpu"] += gpu_usage
            profile_data["total_io"] += io_usage

            # Print immediate profile if duration is above min_duration
            if duration >= min_duration:
                calls = profile_data["calls"]
                avg_time = profile_data["total_time"] / calls
                avg_cpu = profile_data["total_cpu"] / calls
                avg_memory = profile_data["total_memory"] / calls
                avg_gpu = profile_data["total_gpu"] / calls
                avg_io = profile_data["total_io"] / calls
                notes = profile_data.get("notes", "")

                display_profile_info(
                    name=name,
                    duration=duration,
                    cpu_usage=cpu_usage,
                    mem_usage=mem_usage,
                    gpu_usage=gpu_usage,
                    io_usage=io_usage,
                    avg_time=avg_time,
                    avg_cpu=avg_cpu,
                    avg_memory=avg_memory,
                    avg_gpu=avg_gpu,
                    avg_io=avg_io,
                    calls=calls,
                    notes=notes,
                    quiet=quiet,
                    min_duration=min_duration
                )
